'''# pyncm
本模块提供`Get/Set/DumpCurrentSession`,`LoadNewSessionFromDump` 以管理请求
API 使用请参见 `pyncm.apis` 文档'''

from .utils.crypto import Crypto
from typing import Text, Union
import requests,logging,time,json

class Session(requests.Session):   
    '''Represents an API session

    Args:
        requests ([type]): [description]

    Returns:
        [type]: [description]
    '''
    HOST = "https://music.163.com"
    
    def __init__(self,*a,**k):
        # Setting up our request session
        super().__init__(*a,**k)
        self.headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "User-Agent": "mos9527 Him Self/v15",
            "Referer":Session.HOST
        }
        # Setting up default values
        self.login_info = {'success': False,'tick': time.time(), 'content': None}
        self.csrf_token = ''


    def request(self, method: str, url: Union[str, bytes, Text], *a,**k) -> requests.Response:
        '''Initiates & fires a request

        Args:
            method (str): [HTTP Verb]
            url (Union[str, bytes, Text]): [Full / partial API URL]

        Returns:
            requests.Response: [A requests.Response object]
        '''
        if url[:4] != 'http':url = Session.HOST + url
        return super().request(method,url,*a,**k)             
    
    DUMPS = {
        # 'name':(getfunc,setfunc)
        'login_info':(lambda self:getattr(self,'login_info'),lambda self,v:setattr(self,'login_info',v)),
        'csrf_token':(lambda self:getattr(self,'csrf_token'),lambda self,v:setattr(self,'csrf_token',v)),
        'cookies':   (lambda self:getattr(self,'cookies').get_dict(),lambda self,v:getattr(self,'cookies').update(v))
    }
    def dump(self):
        return {name:Session.DUMPS[name][0](self) for name in Session.DUMPS.keys()}
        
    def load(self,dumped):
        for k,v in dumped.items():Session.DUMPS[k][1](self,v)
        return True

class SessionManager():
    '''Manages session switching / loading / dumping'''
    def __init__(self) -> None:
        self.session = Session()
    def get(self):
        return self.session
    def set(self,session):
        self.session = session
    def dump(self):
        return Crypto.EapiCrypto('pyncm',json.dumps(self.session.dump()))['params']
    def load(self,dump : str):
        self.session = Session() # a new session WILL be created
        dump = Crypto.HexCompose(dump)
        dump = Crypto.EapiDecrypt(dump).decode()
        dump = dump.split('-36cd479b6b5-')
        assert dump[0] == 'pyncm'
        return self.session.load(json.loads(dump[1]))
       
logger = logging.getLogger('ncm')
'''Logger to be used by local modules'''
sessionManager          = SessionManager()
'''Manages active session'''
GetCurrentSession       = sessionManager.get
'''Retrives current active session'''
SetCurrentSession       = sessionManager.set
'''Sets current active session'''
DumpCurrentSession      = sessionManager.dump
'''Dumps current session as encrypted hex string'''
LoadNewSessionFromDump  = sessionManager.load
'''Loads session from file'''
from .apis import *


'''# 网易云音乐 APIs'''
import base64
from time import time
from requests.api import head, request
from .. import GetCurrentSession, Session
from .. import logger
from .. import Crypto
import json,zlib,urllib.parse

class LoginRequiredException(Exception):pass
class LoginFailedException(Exception):pass
# region Base models export
LOGIN_REQUIRED = LoginRequiredException('Function needs you to be logged in')

def LoginRequiredApi(func):
    '''This API needs the user's token'''
    def wrapper(*a,**k):
        if not GetCurrentSession().login_info['success']:raise LOGIN_REQUIRED
        return func(*a,**k)
    return wrapper

def UserIDBasedApi(func):
    '''This API has UID as its first argument'''
    def wrapper(user_id=0,*a,**k):
        if user_id == 0 and GetCurrentSession().login_info['success']:
            user_id = GetCurrentSession().login_info['content']['account']['id']
        elif user_id == 0:
            raise LOGIN_REQUIRED
        return func(user_id,*a,**k)
    return wrapper

def _BaseWrapper(requestFunc):
    '''Base wrapper for crypto requests
    '''
    def apiWrapper(apiFunc):
        '''apiFunc -> rtype : (url,plain,[method])'''
        def wrapper(*a,**k):
            ret       = apiFunc(*a,**k)
            url,plain = ret[:2]
            method    = ret[-1] if ret[-1] in ['POST','GET'] else 'POST'        
            rsp = requestFunc(url,plain,method)            
            try:
                return json.loads(rsp.text)
            except json.JSONDecodeError:
                return rsp
        return wrapper
    return apiWrapper

@_BaseWrapper
def WeapiCryptoRequest(url,plain,method):
    '''This apis utilize `/weapi/` calls,seen in web & pc clients'''
    payload = json.dumps({**plain,'csrf_token':GetCurrentSession().csrf_token})
    return GetCurrentSession().request(method,
        GetCurrentSession().HOST + url.replace('/api/','/weapi/'),
        params={'csrf_token':GetCurrentSession().csrf_token},
        data={**Crypto.WeapiCrypto(payload)},
    )
# region Port of `Binaryify/NeteaseCloudMusicApi`
@_BaseWrapper
def LapiCryptoRequest(url,plain,method):
    '''This api proxies over /api/linux/forward,seen in Desktop Linux based clients'''
    payload = {
        'method':method,
        'url':GetCurrentSession().HOST + url,
        'params':plain
    }
    payload = json.dumps(payload)
    return GetCurrentSession().request(method,
        '/api/linux/forward',
        headers={
            **GetCurrentSession().headers, 
            'User-Agent':'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36',
        },
        data={**Crypto.LinuxCrypto(payload)}
    )

@_BaseWrapper
def EapiCryptoRequest(url,plain,method):
    '''This API utilize '/api/' or '/eapi/',seen in mobile clients'''
    cookies = {
        'appver': '6.1.1',
        'buildver':'undefined',   
        'channel':'undefined',
        'deviceId': 'undefined',             
        'mobilename' : 'undefined',        
        'os': 'android',
        'osver':'undefined',            
        'requestId':f'{int(time() * 1000)}_0233',
        'resolution': '1920x1080',        
        'versioncode': '140',
        '__csrf':GetCurrentSession().csrf_token
    }
    payload = {
        **plain,
        'header':cookies
    }
    request = GetCurrentSession().request(method,
        GetCurrentSession().HOST + url,
        headers={
            **GetCurrentSession().headers, 
            'User-Agent':'Mozilla/5.0 (Linux; Android 10; Pixel 2 XL) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Mobile Safari/537.36',
        },
        cookies=cookies,
        data={**Crypto.EapiCrypto(url.replace('/eapi/','/api/'),json.dumps(payload))}
    )
    return request
# endregion
# endregion
from . import miniprograms,album,cloud,cloudsearch,login,playlist,track,user,video


'''网易云搜索 - Cloudsearch APIs'''
from . import WeapiCryptoRequest
TYPE_SONG     = 1
TYPE_ALBUM    = 10
TYPE_ARTIST   = 100
TYPE_PLAYLIST = 1000
TYPE_USER     = 1002
TYPE_MV       = 1004
TYPE_LYRICS   = 1006
TYPE_DJ       = 1009
TYPE_VIDEO    = 1014
@WeapiCryptoRequest
def Search(keyword:str,type=TYPE_SONG,limit=30,offset=0):
    '''搜索某类型关键字

    Args:
        keyword (str): 搜索关键字
        type ([int, optional): 搜索类型，请使用本模块内 TYPE_... 之一 Defaults to TYPE_SONG.
        limit (int, optional): 单次获取量. Defaults to 30.
        offset (int, optional): 获取偏移数. Defaults to 0.

    Returns:
        dict
    '''
    return '/weapi/cloudsearch/get/web',{'s':str(keyword),'type':str(type),'limit':str(limit),'offset':str(offset)}


'''歌单 - Playlist APIs'''
from . import LapiCryptoRequest, WeapiCryptoRequest

@WeapiCryptoRequest
def GetPlaylistInfo(playlist_id,offset=0,total=True,limit=1000):
    '''获取歌单内容

    Args:
        playlist_id ([type]): 歌单 ID
        offset (int, optional): 获取偏移数. Defaults to 0.
        total (bool, optional): 是否获取全部内容. Defaults to True.
        limit (int, optional): 单次获取量. Defaults to 30.
    
    Returns:
        dict
    '''
    return '/weapi/v6/playlist/detail',{"id":str(playlist_id),"offset":str(offset),"total":str(total).lower(),"limit":str(limit)}

@WeapiCryptoRequest
def GetPlaylistComments(playlist_id : str, offset=0, limit=20 ,beforeTime=0):
    '''获取歌单评论

    Args:
        album_id (str): 歌单ID
        offset (int, optional): 时间顺序偏移数. Defaults to 0.
        limit (int, optional): 单次评论获取量. Defaults to 20.
        beforeTime (int, optional): 评论将从该时间戳（秒为单位）开始. Defaults to 0.

    Returns:
        dict
    '''
    return '/v1/resource/comments/A_PL_0_%s' % playlist_id,{'rid':str(playlist_id),'limit':str(limit),'offset':str(offset),'beforeTime':str(beforeTime * 1000)}

@LapiCryptoRequest
def GetTopPlaylists():
    '''获取热门歌单

    Returns:
        dict
    '''
    return '/api/toplist',{}



'''登录、CSRF 有关 APIs'''
from . import EapiCryptoRequest, WeapiCryptoRequest,GetCurrentSession,logger,Crypto,LoginFailedException
import time
def LoginViaCellphone(phone='', password='',remeberLogin=True) -> dict:
    '''网页端 - 手机号登陆

    Args:
        phone (str, optional): 手机号. Defaults to ''.
        password (str, optional): 明文密码. Defaults to ''.
        remeberLogin (bool, optional): 是否‘自动登录’，开启可延长 CSRF 令牌时效. Defaults to True.

    Raises:
        LoginFailedException: 当登陆失败时发生

    Returns:
        dict
    '''
    sess = GetCurrentSession()
    if (phone and password):
        sess.phone = phone
        sess.password = password
        LoginViaCellphone()
    else:
        if (sess.phone and sess.password):
            md5_password = Crypto.HashDigest(sess.password)
            r = WeapiCryptoRequest(lambda:('/weapi/login/cellphone',{"phone":str(sess.phone),"password":str(md5_password),"rememberLogin":str(remeberLogin).lower(),"checkToken":""}))()
            sess.login_info = {'tick': time.time(), 'content':r}
            if not sess.login_info['content']['code'] == 200:
                sess.login_info['success'] = False
                raise LoginFailedException(sess.login_info['content'])
            sess.login_info['success'] = True
            cookie = sess.cookies.get_dict()
            sess.csrf_token = cookie['__csrf']
            logger.debug('Updated login info for user %s' % sess.login_info['content']['profile']['nickname'])
            return sess.login_info
        else:
            raise LoginFailedException('Not enough login info provided')

@EapiCryptoRequest
def CheckIsCellphoneRegistered(cell : str,prefix=86):
    '''检查某手机号是否已注册

    Args:
        cell (str): 手机号
        prefix (int): 区号 . Defaults to 86

    Returns:
        dict
    '''
    return '/eapi/cellphone/existence/check',{'cellphone':cell,'countrycode':prefix}